%{
/*
 * tcng.y - New TC configuration language
 *
 * Written 2001 by Werner Almesberger
 * Copyright 2001 EPFL-ICA, Network Robots
 */

/*
 * Note: this parser is full of memory leaks. It makes it much easier to
 * implement, and isn't a problem in real life, because it only runs for
 * one translation pass, and then all memory is freed on process exit.
 */


#include <stdint.h>
#include <stddef.h> /* for NULL */
#include <stdio.h>
#include <stdlib.h>

#include "config.h"
#include "tcdefs.h"
#include "util.h"
#include "var.h"
#include "data.h"
#include "op.h"
#include "param.h"
#include "tree.h"
#include "field.h"
#include "named.h"
#include "device.h"
#include "qdisc.h"
#include "filter.h"
#include "police.h"


static ELEMENT *current_element = NULL;
static PARAM_DEF *param_def = NULL;
static uint32_t *number_anchor = NULL;

QDISC *root = NULL;
static PARENT parent;
static PARENT tmp_parent = { .qdisc = NULL };


#define NONZERO(dd) \
  if (!(dd->data.type == dt_unum ? dd->data.u.unum : dd->data.u.fnum)) \
    yyerrorf("parameter \"%s\" must be non-zero",dd->dsc->id);
#define UNUM_RANGE(dd,fmt,lower,upper) \
  if (dd->data.u.unum < lower) \
    yyerrorf("parameter \"%s\" value " fmt " below limit " fmt, \
      dd->dsc->id,(unsigned long) dd->data.u.unum,(unsigned long) lower); \
  if (dd->data.u.unum > upper) \
    yyerrorf("parameter \"%s\" value " fmt " above limit " fmt, \
      dd->dsc->id,(unsigned long) dd->data.u.unum,(unsigned long) upper)


static void create_device(const char *name)
{
    DEVICE *device;

    device = alloc_t(DEVICE);
    device->name = name;
    device->qdisc = NULL;
    device->ingress = NULL;
    device->next = NULL;
    add_device(device);
    parent.device = device;
    parent.class = NULL;
    parent.filter = NULL;
    param_def = NULL;
    number_anchor = NULL;
}


static CLASS *begin_class(void)
{
    CLASS *class;

    if (!parent.qdisc->dsc->class_param)
	yyerrorf("qdisc %s has no classes",parent.qdisc->dsc->name);
    class = alloc_t(CLASS);
    class->parent = parent;
    add_class(class);
    class->number = UNDEF_U32;
    class->filters = NULL;
    class->child = NULL;
    class->next = NULL;
    parent.class = class;
    param_def = parent.qdisc->dsc->class_param;
    number_anchor = &class->number;
    var_begin_level();
    return class;
}


static void end_class(PARAM *params,QDISC *qdisc)
{
    var_end_level();
    parent.class->params = params;
    check_required(params,parent.qdisc->dsc->class_param->required);
    parent.class->qdisc = qdisc;
    parent = parent.class->parent;
}


static void begin_element(void)
{
    if (!parent.filter) yyerror("filter element without filter");
    if (!parent.filter->dsc->element_param)
	yyerrorf("filter %s has no elements",parent.filter->dsc->name);
    if (!parent.class) yyerror("filter element without class");
    current_element = alloc_t(ELEMENT);
    add_element(current_element,parent);
    current_element->number = UNDEF_U32;
    current_element->police = NULL;
    current_element->next = NULL;
    param_def = parent.filter->dsc->element_param;
    number_anchor = &current_element->number;
}


static void end_element(PARAM *params,POLICE *police)
{
    current_element->params = params;
    check_required(params,parent.filter->dsc->element_param->required);
    current_element->police = police; 
    current_element = NULL;
}


%}

%union {
    int unum;
    double fnum;
    const char *str;
    DATA data;
    DATA_LIST *list;
    VAR *var;
    DEVICE *dev;
    QDISC *qdisc;
    QDISC_DSC *qddsc;
    CLASS *class;
    FILTER *filter;
    FILTER_DSC *fdsc;
    ELEMENT *element;
    POLICE *police;
    DECISION decis;
    PARAM *param;
    FIELD *field;
    struct {
	DATA b,c;
    } mask;
};


%token		TOK_QDISC TOK_CLASS TOK_FILTER TOK_ELEMENT TOK_TUNNEL TOK_ON
%token		TOK_POLICE TOK_ELSE TOK_IF TOK_IF_ANCHOR TOK_HOST
%token		TOK_PASS TOK_RECLASSIFY TOK_DROP TOK_CONTINUE TOK_FIELD
%token		TOK_CONFORM TOK_COUNT
%token		TOK_CBQ TOK_DSMARK TOK_FIFO TOK_GRED TOK_PRIO TOK_RED TOK_SFQ
%token		TOK_TBF TOK_INGRESS
%token		TOK_FW TOK_ROUTE TOK_RSVP TOK_TCINDEX
%token		TOK_AH TOK_ALLOT TOK_AVPKT TOK_BANDS TOK_BANDWIDTH TOK_BOUNDED
%token		TOK_BURST
%token		TOK_DEFAULT TOK_DEFAULT_INDEX TOK_DPORT TOK_DST
%token		TOK_ECN TOK_ESP TOK_EWMA TOK_FALL_THROUGH
%token		TOK_FROM TOK_FROMIF TOK_GRIO TOK_HASH
%token		TOK_INDICES TOK_ISOLATED TOK_IPPROTO
%token		TOK_LIMIT TOK_MASK TOK_MAX TOK_MAXBURST TOK_MIN TOK_MINBURST
%token		TOK_MPU TOK_MTU TOK_ORDER TOK_OVERFLOW
%token		TOK_PASS_ON TOK_PEAKRATE TOK_PERTURB TOK_PRIOMAP
%token		TOK_PROBABILITY TOK_PROTOCOL TOK_QUANTUM TOK_RATE
%token		TOK_SET_TC_INDEX TOK_SHIFT TOK_SKIP TOK_SPORT TOK_SRC
%token		TOK_TO TOK_VALUE TOK_WEIGHT
%token		SHIFT_RIGHT SHIFT_LEFT LOGICAL_AND LOGICAL_OR
%token		REL_EQ REL_NE REL_LE REL_GE

%token <unum>	UNUM SIZE_QUAL IPV4
%token <fnum>	FNUM
%token <str>	WORD FIELD_NAME VARIABLE
%token <data>	UNIT_SUFFIX

%type  <str>	device
%type  <qdisc>	qdisc_spec opt_class_body class_body class_body_elements
%type  <qddsc>	qdisc_name
%type  <class>	class_spec
%type  <filter>	filter_spec opt_filter
%type  <police>	opt_police police_spec
%type  <decis>	opt_else policing_decision opt_policing_decision
%type  <fdsc>	filter_name
%type  <param>	opt_parameters parameters parameter_list parameter
%type  <param>  parameter_expression
%type  <data>	opt_condition
%type  <data>	complex_expression constant_expression
%type  <data>	mask_expression unary_expression simple_expression
%type  <data>	additive_expression shift_expression multiplicative_expression
%type  <data>	and_expression inclusive_or_expression exclusive_or_expression
%type  <data>	relational_expression equality_expression expression
%type  <data>	logical_and_expression logical_or_expression
%type  <data>	qdisc_expression class_expression filter_expression
%type  <data>	police_expression
%type  <data>	primary_expression variable_expression
%type  <data>	opt_unit_qualifier opt_mask_len
%type  <data>	access_expression opt_index_expression
%type  <data>	host host_word port ipproto ether_proto
%type  <list>	forward_class_list
%type  <unum>	opt_size_qualifier
%type  <mask>	mask_selector
%type  <var>	variable

%%


all:
	{
	    field_set("raw",op_ternary(&op_access,data_none(),data_none(),
	      data_unum(0)),data_none());
	}
    fields assignments opt_no_dev_specs device_specs
    ;


/* -------------------------------- Fields --------------------------------- */


fields:
    | field fields
    ;

field:
    TOK_FIELD WORD '=' expression opt_condition
	{
	    field_set($2,$4,$5);
	    free($2);
	}
    | TOK_FIELD FIELD_NAME '=' expression opt_condition
	{
	    field_redefine($2,$4,$5);
	}
    ;

opt_condition:
	{
	    $$ = data_none();
	}
    | TOK_IF expression
	{
	    $$ = $2;
	}
    | TOK_IF_ANCHOR expression /* at this point, we may get either token */
	{
	    $$ = $2;
	}
    ;


/* ------------------------------ Device list ------------------------------ */


device_specs:
    | device_spec device_specs
    ;

device_spec:
	{
	    /* make sure local variables get local scope */
	    var_begin_level();
	    /* @@@ language design bug: if omitting the device, variables
	       get global scope. Possible solutions:
		- require braces for device too, or
		- always require device name (not too bad - can use macros for
		  default anyway; ...
		  ... BUT: would make usage in tcsim a bit ugly)
		 */
	}
    device
	{
	    create_device($2);
	}
      qdisc_specs
	{
	    var_end_level();
	}
    ;

opt_no_dev_specs:
    |
	{
	    create_device(default_device);
		/* stralloc this if ever implementing deallocations */
	}
      qdisc_specs
	{
	    var_end_level(); /* @@@ unmatched ? */
	}
    ;

qdisc_specs:
    single_qdisc_spec
    | single_qdisc_spec qdisc_specs
    ;

single_qdisc_spec:
    qdisc_spec assignments
	{
	    if ($1->dsc == &ingress_dsc) {
		if (parent.device->ingress)
		    yyerror("ingress qdisc is already set");
		parent.device->ingress = $1;
	    }
	    else {
		if (parent.device->qdisc)
		    yyerror("egress qdisc is already set");
		parent.device->qdisc = $1;
	    }
	}
    ;

device:
    WORD no_colon assignments
	{
	    $$ = $1;
	}
    ;

no_colon:
    | ':'
	{
	    warn("colon after interface name is obsolete");
	}
    ;


/* ---------------------- Assignments and expressions ---------------------- */


assignments:
    | assignment assignments
    ;

assignment:
    variable '=' complex_expression
	{
	    var_set($1,$3);
	}
    ;

complex_expression:
    constant_expression
	{
	    $$ = $1;
	}
    | qdisc_expression
	{
	    $$ = $1;
	}
    | class_expression
	{
	    $$ = $1;
	}
    | filter_expression
	{
	    $$ = $1;
	}
    | police_expression
	{
	    $$ = $1;
	}
    ;

qdisc_expression:
    TOK_QDISC variable_expression
	{
	    $$ = data_convert($2,dt_qdisc);
	    if ($$.u.qdisc->parent.device != parent.device)
		yyerror("qdisc was created in different scope");
	}
    | qdisc_spec
	{
	    $$ = data_qdisc($1);
	}
    ;

class_expression:
    TOK_CLASS variable_expression
	{
	    $$ = data_convert($2,dt_class);
	    if ($$.u.class->parent.qdisc != parent.qdisc ||
	      $$.u.class->parent.class != parent.class)
		yyerror("class was created in different scope");
	    $$.u.class->parent = parent;
		/* okay to change it a little - the fields we may modify are
		   only used temporarily during construction anyway */
	    parent.class = $$.u.class;
	    var_begin_level();
	}
      opt_selectors opt_class_body
	{
	    var_end_level();
	    if ($5) {
		if (parent.class->qdisc) yyerror("class already has a qdisc");
		parent.class->qdisc = $5;
	    }
	    parent = parent.class->parent;
	    $$ = $<data>3;
	}
    | class_spec
	{
	    $$ = data_class($1);
	}
    ;

filter_expression:
    TOK_FILTER variable_expression
	{
	    $$ = data_convert($2,dt_filter);
	    if ($$.u.filter->parent.qdisc != parent.qdisc)
		yyerror("filter was created in different scope");
	    $$.u.filter->parent.filter = parent.filter;
	    parent.filter = $$.u.filter;
	    /* we don't dare to touch any other fields of parent */
	    var_begin_level();
	}
      opt_filter_body
	{
	    var_end_level();
	    parent.filter = parent.filter->parent.filter;
	    $$ = $<data>3;
	}
    | filter_spec
	{
	    $$ = data_filter($1);
	    if ($$.u.filter->parent.qdisc != parent.qdisc ||
	      $$.u.filter->parent.class != parent.class)
		yyerror("filter was created in different scope");
	}
    ;

police_expression:
    TOK_POLICE variable_expression
	{
	    $$ = data_convert($2,dt_police);
	}
    | police_spec
	{
	    $$ = data_police($1);
	}
    ;

constant_expression:
    expression
	{
	    $$ = $1;
	    if ($$.op) yyerror("constant expression isn't");
	}
    ;


/*
 * The following expression syntax is an almost exact copy from K&R 2nd ed.,
 * with a few modifications near simple_expression.
 */

expression:
    logical_or_expression
	{
	    $$ = $1;
	}
    ;

logical_or_expression:
    logical_and_expression
	{
	    $$ = $1;
	}
    | logical_or_expression LOGICAL_OR logical_and_expression
	{
	    $$ = op_binary(&op_logical_or,$1,$3);
	}
    ;

logical_and_expression:
    inclusive_or_expression
	{
	    $$ = $1;
	}
    | logical_and_expression LOGICAL_AND inclusive_or_expression
	{
	    $$ = op_binary(&op_logical_and,$1,$3);
	}
    ;

inclusive_or_expression:
    exclusive_or_expression
	{
	    $$ = $1;
	}
    | inclusive_or_expression '|' exclusive_or_expression
	{
	    $$ = op_binary(&op_or,$1,$3);
	}
    ;

exclusive_or_expression:
    and_expression
	{
	    $$ = $1;
	}
    | exclusive_or_expression '^' and_expression
	{
	    $$ = op_binary(&op_xor,$1,$3);
	}
    ;

and_expression:
    equality_expression
	{
	    $$ = $1;
	}
    | and_expression '&' equality_expression
	{
	    $$ = op_binary(&op_and,$1,$3);
	}
    ;

equality_expression:
    relational_expression
	{
	    $$ = $1;
	}
    | equality_expression REL_EQ relational_expression
	{
	    $$ = op_binary(&op_eq,$1,$3);
	}
    | equality_expression REL_NE relational_expression
	{
	    $$ = op_binary(&op_ne,$1,$3);
	}
    ;

relational_expression:
    shift_expression
	{
	    $$ = $1;
	}
    | relational_expression '<' shift_expression
	{
	    $$ = op_binary(&op_lt,$1,$3);
	}
    | relational_expression '>' shift_expression
	{
	    $$ = op_binary(&op_gt,$1,$3);
	}
    | relational_expression REL_LE shift_expression
	{
	    $$ = op_binary(&op_le,$1,$3);
	}
    | relational_expression REL_GE shift_expression
	{
	    $$ = op_binary(&op_ge,$1,$3);
	}
    ;

shift_expression:
    additive_expression
	{
	    $$ = $1;
	}
    | shift_expression SHIFT_RIGHT additive_expression
	{
	    $$ = op_binary(&op_shift_right,$1,$3);
	}
    | shift_expression SHIFT_LEFT additive_expression
	{
	    $$ = op_binary(&op_shift_left,$1,$3);
	}
    ;

additive_expression:
    multiplicative_expression
	{
	    $$ = $1;
	}
    | additive_expression '+' multiplicative_expression
	{
	    $$ = op_binary(&op_plus,$1,$3);
	}
    | additive_expression '-' multiplicative_expression
	{
	    $$ = op_binary(&op_minus,$1,$3);
	}
    ;

multiplicative_expression:
    unary_expression
	{
	    $$ = $1;
	}
    | multiplicative_expression '*' unary_expression
	{
	    $$ = op_binary(&op_mult,$1,$3);
	}
    | multiplicative_expression '/' unary_expression
	{
	    $$ = op_binary(&op_div,$1,$3);
	}
    | multiplicative_expression '%' unary_expression
	{
	    $$ = op_binary(&op_mod,$1,$3);
	}
    ;

unary_expression:
    mask_expression
	{
	    $$ = $1;
	}
    | '!' mask_expression
	{
	    $$ = op_unary(&op_logical_not,$2);
	}
    | '~' mask_expression
	{
	    $$ = op_unary(&op_not,$2);
	}
    ;

mask_expression:
    simple_expression
	{
	    $$ = $1;
	}
    | simple_expression ':' mask_selector
	{
	    $$ = op_ternary(&op_mask,$1,$3.b,$3.c);
	}
/*
 * We don't use mask_expression ':' mask_selector  here, because this would
 * generate an ambiguity for cases like a:b:c:d, which is a little messy to
 * resolve. In any case, multiple masks are hardly of any practical use.
 */
    ;

mask_selector:
    simple_expression opt_mask_len
	{
	    $$.b = $1;
	    $$.c = $2;
	}
    | ':' simple_expression
	{
	    $$.b = data_none();
	    $$.c = $2;
	}
    ;

opt_mask_len:
	{
	    $$ = data_none();
	}
    | ':' simple_expression
	{
	    $$ = $2;
	}
    ;

simple_expression:
    primary_expression opt_unit_qualifier
	{
	    $$ = data_add_unit($1,$2);
	}
    ;

opt_unit_qualifier:
	{
	    $$ = data_none();
	}
    | UNIT_SUFFIX
	{
	    $$ = $1;
	}
    ;

primary_expression:
    variable_expression
	{
	    $$ = $1;
	}
    | UNUM
	{
	    $$ = data_unum($1);
	}
    | IPV4
	{
	    $$ = data_unum($1);
	}
    | FNUM
	{
	    $$ = data_fnum($1);
	}
    | access_expression
	{
	    $$ = $1;
	}
    | TOK_HOST host_word
	{
	    $$ = $2;
	}
    | TOK_HOST IPV4
	{
	    $$ = data_unum($2);
	}
    | '(' expression ')'
	{
	    $$ = $2;
	}
    | TOK_CONFORM police_expression
	{
	    $$ = op_unary(&op_conform,$2);
	}
    | TOK_COUNT police_expression
	{
	    $$ = op_unary(&op_count,$2);
	}
    | TOK_DROP
	{
	    $$ = data_decision(dr_drop,NULL);
	}
    | TOK_RECLASSIFY
	{
	    $$ = data_decision(dr_reclassify,NULL);
	}
    | TOK_PASS
	{
	    static int warn = 1;

	    if (warn) {
		yywarnf("direct use of \"pass\" is discouraged");
		warn = 0;
	    }
	    $$ = data_decision(dr_class,NULL);
	}
    | TOK_CONTINUE
	{
	    static int warn = 1;

	    if (warn) {
		yywarnf("direct use of \"continue\" is discouraged");
		warn = 0;
	    }
	    $$ = data_decision(dr_continue,NULL);
	}
    ;

variable_expression:
    variable
	{
	    $$ = var_get($1);
	}
    ;

variable:
    VARIABLE
	{
	    $$ = var_find($1);
	    free($1);
	}
    ;

access_expression:
    FIELD_NAME opt_index_expression opt_size_qualifier
	{
	    $$ = op_ternary(&op_access,data_field($1),$2,data_unum($3));
	}
    ;

opt_size_qualifier:
	{
	    $$ = 0;
	}
    | '.' UNUM
	{
	    if ($2 != 8 && $2 != 16 && $2 != 32)
		yyerrorf("size must be 8, 16, or 32, not %d",$2);
	    $$ = $2;
	}
    | SIZE_QUAL
	{
	    $$ = $1;
	}
    ;

opt_index_expression:
	{
	    $$ = data_none();
	}
    | '[' expression ']'
	{
	    $$ = $2;
	}
    ;


/* -------------------------- Qdisc specification -------------------------- */


qdisc_spec:
    qdisc_name
	{
	    if (!parent.device) yyerror("qdisc without device");
	    $$ = alloc_t(QDISC);
	    $$->parent = parent;
	    $$->dsc = $1;
	    $$->number = UNDEF_U32;
	    $$->classes = NULL;
	    $$->filters = NULL;
	    $$->if_expr = data_none();
	    parent.qdisc = $$;
	    parent.class = NULL;
	    parent.filter = NULL;
	    param_def = $1->qdisc_param;
	    number_anchor = $1 != &ingress_dsc ? &$$->number : NULL;
	    var_begin_level();
	}
      opt_parameters opt_qdisc_body
	{
	    var_end_level();
	    $$ = $<qdisc>2;
	    $$->params = $3;
	    check_required($3,$$->dsc->qdisc_param->required);
	    parent = $$->parent;
	}
    ;

qdisc_name:
    TOK_INGRESS
	{
	    $$ = &ingress_dsc;
	}
    | TOK_CBQ
	{
	    $$ = &cbq_dsc;
	}
    | TOK_DSMARK
	{
	    $$ = &dsmark_dsc;
	}
    | TOK_FIFO
	{
	    $$ = &fifo_dsc;
	}
    | TOK_GRED
	{
	    $$ = &gred_dsc;
	}
    | TOK_PRIO
	{
	    $$ = &prio_dsc;
	}
    | TOK_RED
	{
	    $$ = &red_dsc;
	}
    | TOK_SFQ
	{
	    $$ = &sfq_dsc;
	}
    | TOK_TBF
	{
	    $$ = &tbf_dsc;
	}
    ;

opt_qdisc_body:
    | '{' qdisc_body '}'
    ;

qdisc_body:
    | qdisc_or_class_body_element qdisc_body
    |
	{
	    begin_class();
	}
      qdisc_expression
	{
	    end_class(NULL,$<data>2.u.qdisc);
	}
      qdisc_body
    ;

qdisc_or_class_body:
    | qdisc_or_class_body_element qdisc_or_class_body
    ;

qdisc_or_class_body_element:
    assignment
    | filter_expression
	{
	    /* ignore value */
	}
    | class_expression
	{
	    /* ignore value */
	}
    ;


/* ------------------------------ Parameters ------------------------------- */


opt_parameters:
	{
	    $$ = NULL;
	}
    | parameters
	{
	    $$ = $1;
	}
    ;

parameters:
    '(' parameter_list ')'
	{
	    $$ = $2;
	}
    ;

parameter_list:
	{
	    $$ = NULL;
	}
    | parameter
	{
	    $$ = $1;
	}
    | parameter ',' parameter_list
	{
	    if ($1) {
		$$ = $1;
		$$->next = $3;
	    }
	    else {
		$$ = $3;
	    }
	}
    ;

parameter:
    parameter_expression
	{
	    $$ = $1;
	    if ($$) check_optional($$,param_def->required,param_def->optional);
	}
    ;

parameter_expression:
    constant_expression
	{
	    if (!number_anchor) yyerror("number/handle not valid here");
	    if (*number_anchor != UNDEF_U32)
		yyerror("number/handle already set");
	    *number_anchor = data_convert($1,dt_unum).u.unum;
	    $$ = NULL;
	}
    | TOK_AH constant_expression
	{
	    $$ = param_make(&prm_ah,$2);
	    NONZERO($$);
	}
    | TOK_ALLOT constant_expression
	{
	    $$ = param_make(&prm_allot,$2);
	    NONZERO($$);
	}
    | TOK_AVPKT constant_expression
	{
	    $$ = param_make(&prm_avpkt,$2);
	    NONZERO($$);
	}
    | TOK_BANDS constant_expression
	{
	    $$ = param_make(&prm_bands,$2);
	    UNUM_RANGE($$,"%lu",TCQ_PRIO_MIN_BANDS,TCQ_PRIO_BANDS);
	}
    | TOK_BANDWIDTH constant_expression
	{
	    $$ = param_make(&prm_bandwidth,$2);
	}
    | TOK_BOUNDED
	{
	    $$ = param_make(&prm_bounded,data_unum(1));
	}
    | TOK_BURST constant_expression
	{
	    $$ = param_make(&prm_burst,$2);
	    /* check against expected psched_mtu */
	}
    | TOK_DEFAULT
	{
	    $$ = param_make(&prm_default,data_unum(1));
	}
    | TOK_DEFAULT_INDEX constant_expression
	{
	    $$ = param_make(&prm_default_index,$2);
	    UNUM_RANGE($$,"0x%lx",0,0xffff);
	}
    | TOK_DPORT port
	{
	    $$ = param_make(&prm_dport,$2);
	    UNUM_RANGE($$,"%lu",1,0xffff);
	}
    | TOK_DST host
	{
	    $$ = param_make(&prm_dst,$2);
	    NONZERO($$);
	}
    | TOK_ECN
	{
	    $$ = param_make(&prm_ecn,data_unum(1));
	}
    | TOK_ESP constant_expression
	{
	    $$ = param_make(&prm_esp,$2);
	    NONZERO($$);
	}
    | TOK_EWMA constant_expression
	{
	    $$ = param_make(&prm_ewma,$2);
	    UNUM_RANGE($$,"%lu",1,TC_CBQ_MAX_EWMA);
	}
    | TOK_FALL_THROUGH
	{
	    $$ = param_make(&prm_fall_through,data_unum(1));
	}
    | TOK_FROM constant_expression
	{
	    $$ = param_make(&prm_from,$2);
	    UNUM_RANGE($$,"0x%lx",0,0xff);
	}
    | TOK_FROMIF WORD
	{
	    $$ = param_make(&prm_fromif,data_string($2));
	    free($2);
	    /* @@@ tcng should check if interface exists */
	}
    | TOK_GRIO
	{
	    $$ = param_make(&prm_grio,data_unum(1));
	}
    | TOK_HASH constant_expression
	{
	    $$ = param_make(&prm_hash,$2);
	    UNUM_RANGE($$,"%lu",1,0x10000);
	}
    | TOK_INDICES constant_expression
	{
	    $$ = param_make(&prm_indices,$2);
	    UNUM_RANGE($$,"0x%lx",1,0x10000);
	    /* 2^n checked in qdisc.c */
	}
    | TOK_IPPROTO ipproto
	{
	    $$ = param_make(&prm_ipproto,$2);
	    UNUM_RANGE($$,"%lu",0,0xff);
	}
    | TOK_ISOLATED
	{
	    $$ = param_make(&prm_isolated,data_unum(1));
	}
    | TOK_LIMIT constant_expression
	{
	    if ($2.type == dt_psize) $$ = param_make(&prm_plimit,$2);
	    else $$ = param_make(&prm_limit,$2);
	}
    | TOK_MASK constant_expression
	{
	    $$ = param_make(&prm_mask,$2);
	    UNUM_RANGE($$,"0x%lx",0,0xffff);
	    /* DSMARK: 0..0xff, checked in qdisc.c */
	}
    | TOK_MAX constant_expression
	{
	    $$ = param_make(&prm_max,$2);
	}
    | TOK_MAXBURST constant_expression
	{
	    $$ = param_make(&prm_maxburst,$2);
	    NONZERO($$);
	}
    | TOK_MIN constant_expression
	{
	    $$ = param_make(&prm_min,$2);
	}
    | TOK_MINBURST constant_expression
	{
	    $$ = param_make(&prm_minburst,$2);
	}
    | TOK_MPU constant_expression
	{
	    $$ = param_make(&prm_mpu,$2);
	}
    | TOK_MTU constant_expression
	{
	    $$ = param_make(&prm_mtu,$2);
	    NONZERO($$);
	}
    | TOK_ORDER constant_expression
	{
	    $$ = param_make(&prm_order,$2);
	    UNUM_RANGE($$,"0x%lx",0,0xff);
	}
    | TOK_OVERFLOW police_expression
	{
	    $$ = param_make(&prm_overflow,$2);
	}
    | TOK_PASS_ON
	{
	    $$ = param_make(&prm_fall_through,data_unum(0));
	}
    | TOK_PEAKRATE constant_expression
	{
	    $$ = param_make(&prm_peakrate,$2);
	}
    | TOK_PERTURB constant_expression
	{
	    $$ = param_make(&prm_perturb,$2);
	}
    | TOK_PRIO constant_expression
	{
	    $$ = param_make(&prm_prio,$2);
	    NONZERO($$);
	}
    | TOK_PRIOMAP forward_class_list
	{
	    $$ = param_make(&prm_priomap,data_list($2));
	}
    | TOK_PROBABILITY constant_expression
	{
	    $$ = param_make(&prm_probability,$2);
	}
    | TOK_PROTOCOL ether_proto
	{
	    $$ = param_make(&prm_protocol,$2);
	    UNUM_RANGE($$,"0x%lx",0,0xffff);
	}
    | TOK_QUANTUM constant_expression
	{
	    $$ = param_make(&prm_quantum,$2);
	    NONZERO($$);
	}
    | TOK_RATE constant_expression
	{
	    $$ = param_make(&prm_rate,$2);
	    /* NONZERO($$); @@@ policer may have zero rate */
	}
    | TOK_SET_TC_INDEX
	{
	    $$ = param_make(&prm_set_tc_index,data_unum(1));
	}
    | TOK_SHIFT constant_expression
	{
	    $$ = param_make(&prm_shift,$2);
	    UNUM_RANGE($$,"%lu",0,15);
	}
    | TOK_SKIP constant_expression
	{
	    $$ = param_make(&prm_skip,$2);
	    NONZERO($$);
	}
    | TOK_SPORT port
	{
	    $$ = param_make(&prm_sport,$2);
	    UNUM_RANGE($$,"%lu",1,0xffff);
	}
    | TOK_SRC host
	{
	    $$ = param_make(&prm_src,$2);
	    NONZERO($$);
	}
    | TOK_TO constant_expression
	{
	    $$ = param_make(&prm_to,$2);
	    UNUM_RANGE($$,"0x%lx",0,0xff);
	}
    | TOK_VALUE constant_expression
	{
	    $$ = param_make(&prm_value,$2);
	    UNUM_RANGE($$,"0x%lx",0,0xff);
	}
    | TOK_WEIGHT constant_expression
	{
	    $$ = param_make(&prm_weight,$2);
	    NONZERO($$);
	}
    ;

host:
    constant_expression
	{
	    $$ = $1;
	}
    | host_word
	{
	    $$ = $1;
	}
    ;

host_word:
    WORD
	{
	    $$ = data_unum(ipv4_host($1));
	    free($1);
	}
    ;

port:
    constant_expression
	{
	    $$ = $1;
	}
    | WORD
	{
	    $$ = data_unum(ip_port($1));
	    free($1);
	}
    ;

ipproto:
    constant_expression
	{
	    $$ = $1;
	}
    | WORD
	{
	    $$ = data_unum(ip_proto($1));
	    free($1);
	}
    ;

ether_proto:
    constant_expression
	{
	    $$ = $1;
	}
    | WORD
	{
	    $$ = data_unum(ether_proto($1));
	    free($1);
	}
    ;

/*
 * Lists are linked lists of DATA_LIST elements, rooted at a DATA of type
 * dt_list. The actual list content is accessed via d.u.list-{>next-}>ref.
 *
 * Magic classes are in fact forward-references. We use the variables before
 * assigning to them. This may look a little odd, but at least the semantics
 * are pretty bullet-proof.
 *
 * Something slightly ugly about forward-references: we need to back-patch the
 * storage location, so it can't be copied by value. Fortunately, lists just
 * happen to copy their elements by reference. What a lucky coincidence :-)
 */

forward_class_list:
	{
	    $$ = NULL;
	}
    | variable forward_class_list
	{
	    $$ = data_list_element(var_forward($1));
	    $$->next = $2;
	}
    ;


/* -------------------------------- Classes -------------------------------- */


/*
 * In principle, this is
 * TOK_CLASS opt_parameters opt_selectors opt_class_body
 * be we disallow the case of TOK_CLASS followed by nothing else, because
 * this would conflict with variable assignments.
 */

class_spec:
    TOK_CLASS
	{
	    $$ = begin_class();
	}
      parameters opt_selectors opt_class_body
	{
	    $$ = $<class>2;
	    end_class($3,$<qdisc>5);
	}
    | TOK_CLASS
	{
	    $$ = begin_class();
	}
      selectors opt_class_body
	{
	    $$ = $<class>2;
	    end_class(NULL,$<qdisc>4);
	}
    | TOK_CLASS
	{
	    $$ = begin_class();
	}
      class_body
	{
	    $$ = $<class>2;
	    end_class(NULL,$<qdisc>3);
	}
    ;


opt_class_body:
	{
	    $$ = NULL;
	}
    | class_body
	{
	    $$ = $1;
	}
    ;

class_body:
    '{' class_body_elements '}'
	{
	    $$ = $2;
	}
    ;

class_body_elements:
	{
	    $$ = NULL;
	}
    | qdisc_or_class_body_element class_body_elements
	{
	    $$ = $2;
	}
    | qdisc_expression qdisc_or_class_body
	{
	    $$ = $1.u.qdisc;
	}
    ;

opt_selectors:
	{
	    if (parent.filter) yywarn("class without selector inside filter");
	}
    | selectors
    ;

selectors:
    selector
    | selector selectors
    ;

selector:
    TOK_ON on_selector
    | TOK_IF opt_filter expression
	{
	    DATA d = $3;

	    debug_expr("Raw IF expression",d);
	    d = field_expand(d);
	    debug_expr("Expanded IF expression",d);
	    if ($2 && $2->dsc != &if_dsc)
		yyerrorf("if requires filter \"if\", not \"%s\"",$2->dsc->name);
	    add_if(d,parent,$2);
	}
    ;

on_selector:
	{
	    /*
	     * This is a little ugly. Background:
	     *  - element gets class from parent.class
	     *  - if filter_expression generates a filter, its parent is the
	     *    parent of the current class.
	     */
	    /* @@@ make it work ? */
	    tmp_parent = parent;
	    parent = parent.class->parent;
	}
    filter_expression
	{
	    parent = tmp_parent;
	    parent.filter = $2.u.filter;
	    if (parent.class == &class_is_tunnel &&
	      parent.filter != parent.tunnel->parent.filter)
		yyerror("can't assign tunnel to different filter");
	    parent.tunnel = NULL;
	}
      element_spec
	{
	    parent = tmp_parent;
	}
    | variable_expression
	{
	    tmp_parent = parent;
	    parent.filter = data_convert($1,dt_filter).u.filter;
	    if (parent.filter->parent.qdisc != parent.qdisc)
		yyerror("filter was created in different scope");
	    if (parent.class == &class_is_tunnel &&
	      parent.filter != parent.tunnel->parent.filter)
		yyerror("can't assign tunnel to different filter");
	    parent.tunnel = NULL;
	    begin_element();
	}
      parameters opt_police
	{
	    end_element($3,$4);
	    parent = tmp_parent;
	}
    | 
	{
	    begin_element();
	}
      parameters opt_police
	{
	    end_element($2,$3);
	}
    | element_spec
    ;

opt_filter:
	{
	    $$ = NULL;
	}
    | TOK_FILTER variable_expression
	{
	    $$ = data_convert($2,dt_filter).u.filter;
	    if ($$->parent.qdisc != parent.qdisc)
		yyerror("filter was created in different scope");
	}
    ;


/* -------------------------------- Filters -------------------------------- */


filter_spec:
    filter_name
	{
	    $$ = alloc_t(FILTER);
	    $$->parent = parent;
	    add_filter($$);
	    $$->number = UNDEF_U32;
	    $$->dsc = $1;
	    $$->elements = NULL;
	    $$->next = NULL;
	    param_def = $1->filter_param;
	    number_anchor = &$$->number;
	    parent.filter = $$;
	    var_begin_level();
	}
	  opt_parameters opt_filter_body
	{
	    var_end_level();
	    $$ = $<filter>2;
	    $$->params = $3;
	    check_required($3,$$->dsc->filter_param->required);
	    parent = $$->parent;
	}
    ;

filter_name:
    TOK_FW
	{
	    $$ = &fw_dsc;
	}
    | TOK_ROUTE
	{
	    $$ = &route_dsc;
	}
    | TOK_RSVP
	{
	    $$ = &rsvp_dsc;
	}
    | TOK_TCINDEX
	{
	    $$ = &tcindex_dsc;
	}
    | TOK_IF_ANCHOR
	{
	    yyerror("\"if\" anchor doesn't work yet and will be re-designed");
	    $$ = &if_dsc;
	}
    ;

opt_filter_body:
    | '{' filter_body '}'
    ;

filter_body:
    | filter_body_element filter_body
    ;

filter_body_element:
    assignment
    | class_expression
	{
	    /* ignore value */
	}
    | tunnel
    ;


/* -------------------------------- Tunnels -------------------------------- */


tunnel:
    TOK_TUNNEL
	{
	    TUNNEL *tunnel;

	    tunnel = alloc_t(TUNNEL);
	    tunnel->parent = parent;
	    tunnel->number = UNDEF_U32;
	    parent.tunnel = tunnel;
	    parent.class = &class_is_tunnel;
	    param_def = &tunnel_param;
	    number_anchor = &tunnel->number;
	    var_begin_level();
	}
      opt_parameters opt_selectors
	{
	    parent.class = parent.tunnel->parent.class;
	}
      opt_filter_body
	{
	    var_end_level();
	    parent.tunnel->params = $3;
	    check_required($3,tunnel_param.required);
	    parent = parent.tunnel->parent;
	}
    ;

/* ---------------------------- Filter Elements ---------------------------- */


element_spec:
    TOK_ELEMENT
	{
	    begin_element();
	}
      opt_parameters opt_police
	{
	    end_element($3,$4);
	}
    ;

opt_police:
	{
	    $$ = NULL;
	}
    | police_expression
	{
	    $$ = $1.u.police;
	}
    ;

/*
 * Not opt_parameters, because a) this would conflict with variable assignment,
 * and b) it wouldn't make sense anyway.
 */

police_spec:
    TOK_POLICE
	{
	    $$ = alloc_t(POLICE);
	    $$->number = UNDEF_U32;
	    number_anchor = &$$->number;
	    param_def = &police_def;
	}
      parameters opt_policing_decision opt_else
	{
	    $$ = $<police>2;
	    $$->params = $3;
	    $$->out_profile = $4;
	    $$->in_profile = $5;
	    check_required($3,police_def.required);
	    add_police($$);
	}
    ;

opt_policing_decision:
	{
	    $$ = pd_reclassify;
	}
    | policing_decision
	{
	    $$ = $1;
	}
    ;

opt_else:
	{
	    $$ = pd_ok;
	}
    | TOK_ELSE policing_decision
	{
	    $$ = $2;
	}
    ;

policing_decision:
    TOK_PASS
	{
	    $$ = pd_ok;
	}
    | TOK_RECLASSIFY
	{
	    $$ = pd_reclassify;
	}
    | TOK_DROP
	{
	    $$ = pd_drop;
	}
    | TOK_CONTINUE
	{
	    $$ = pd_continue;
	}
    ;
